;; The first three lines of this file were inserted by DrScheme. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "reader.ss" "plai" "lang")
;; evt.ss
;;
;; Implementation of environments for the Eph language
;;
;; CS334: Principles of Programming Languages
;; http://cs.williams.edu/~morgan/www/cs334


; bindings is a hash table (hasheq) mapping identifiers to
; VALs. parent is either empty (for the root) or is the parent
; environment.
(define-struct evt (parent bindings))

; Creates an empty environment with a parent
(define (make-child-evt parent)
  (make-evt parent (make-hasheq)))

; Creates the root environment (no parent)
(define (make-root-evt) (make-child-evt empty)) 
   
; Used by environments to create error messages
(define (variable-error msg id)
  (error 'Eph (string-append msg (symbol->string id))))

; Used internally to represent uninitialized values.  Note that this
; is a Scheme symbol, not an Eph symbol.
(define UNINITIALIZED-VAR 'uninitialized-var)


; (evt-get evt id)
; (evt-get evt id b)  -- if b = #f, returns UNINITIALIZED-VAR instead of creating an error on 
;                        uninitialized variables
;
; Creates an error if not found, or uninitialized (i.e., in the middle
; of a DEFINE statement). Recurses into parents.
(define (evt-get evt id . error-on-uninit)
  ; Get the id's value, or #f if it is not bound in this environment
  (let ([value (hash-ref (evt-bindings evt) id #f)])
    (if value
        ; Found in this environment
        (if (eq? value UNINITIALIZED-VAR)
            ; Tried to access an uninitialized variable
            (if (or (empty? error-on-uninit) (not (first error-on-uninit)))
                (variable-error "Uninitialized variable: " id)
                ; Hide the error
                UNINITIALIZED-VAR)
            
            ; Value is initialized; return it
            value)
        
        ; Not found; look in parent environment
        (let ([parent (evt-parent evt)])
          (if (null? parent)
              ; There is no parent!
              (variable-error "Variable not found: " id)
              
              ; Recurse into parent
              (evt-get parent id))))))


; "not eq?" helper
(define (neq? a b) (not (eq? a b)))

; Create a new binding in this environment for id in this environment
(define (evt-define! evt id value)
  (if (evt-this-contains? evt id)
      (error 'Eph (string-append "Cannot define the same variable " 
                                 (symbol->string id) " twice"))
      (hash-set! (evt-bindings evt) id value)))

; Change existing binding for id to value in evt, recursing into parent environments
; if necessary to find the existing binding.
(define (evt-set! evt id value)
  (if (evt-this-contains? evt id)
      
      (hash-set! (evt-bindings evt) id value)

      (let ([parent (evt-parent evt)])
        (if (empty? parent)
            ; Hit the root environment without finding 
            (error 'Eph (string-append "Cannot set! variable "
                                       (symbol->string id)
                                       " before it is defined"))
            ; Look in parent
            (evt-set! parent id value)))))

; Returns non-false if id is initialized. Recurses into parents
(define (evt-initialized? evt id)
  (neq? UNINITIALIZED-VAR (evt-get evt id #f)))

; Does not recurse into parent environments
(define (evt-this-contains? evt id)
  ; Since the environment may only contain VAL and Scheme #f is not
  ; in VAL, if hash-ref returns #f, then we know that id was not
  ; present
  (neq? #f (hash-ref (evt-bindings evt) id #f)))
