;; The first three lines of this file were inserted by DrScheme. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "reader.ss" "plai" "lang")
;; grammar.ss
;;
;; Expression and Value domain types for the Eph language
;;
;; CS334: Principles of Programming Languages
;; http://cs.williams.edu/~morgan/www/cs334

(require "evt.ss")

;; BNF for the Eph language
;;
;; <EXPR>     ::= <LIT> | <LAMBDA> | <IF> | <AND> | <OR> | 
;;                <VAR> | <APP> | <BEGIN> | <DEFINE>
;;
;;    <LIT>   ::= <NUM> | <BOOL> | <SYM> | "null" | "void"
;;      <NUM> ::= <Scheme number>
;;      <BOOL>::= <Scheme boolean: #t #f true false>
;;      <SYM> ::= "(" "quote" <Scheme symbol> ")"
;;    <IF>    ::= "(" "if" <EXPR> <EXPR> <EXPR> ")"
;;    <AND>   ::= "(" "and" <EXPR> <EXPR> ")"
;;    <OR>    ::= "(" "or" <EXPR> <EXPR> ")"
;;    <VAR>   ::= <ID>
;;      <ID>  ::= <Scheme symbol that is not reserved>
;;    <DEFINE>::= "(" "define" <VAR> <EXPR> "]")+ ")" <EXPR> ")"
;;    <APP>   ::= "(" <EXPR>+ ")"
;;    <SET!>  ::= "(" "set!" <ID> <EXPR> ")"
;;    <BEGIN> ::= "(" "begin" <EXPR>* ")"
;;    <LAMBDA>::= "(" ("lambda" | "λ") "(" <ID>* ")" <EXPR> ")"
;;
;;
;; <VAL>      ::= <NUM> | <BOOL> | <SYM> | <PROC> | 
;;                <PAIR> | <VOID> | <NULL> | <BUILTIN-PROC>
;;
;; BUILTIN-PROC uses a Scheme procedure to apply "native"
;; functionality

; These are not allowed to be identifiers
(define reserved-word? 
  (let ([illegal '(quote define set! lambda λ null void and true false with or if begin)])
    (lambda (x) (memq x illegal))))
                               
; True if x is a legal identifier in Eph
(define (ID? x) (and (symbol? x) (not (reserved-word? x))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; Simple Scheme value
(define (simple-value? x)
  (or (number? x) (symbol? x) (boolean? x) (null? x) (void? x)))


; Implementation of the expression domain
; <AND> and <OR> are rewritten by the parser
(define-type EXPR
  (LIT    (v simple-value?))
  
  (IF     (test EXPR?) (consequent EXPR?) (alternate EXPR?))

  (VAR    (name ID?))
  
  ; Subexprs is a list of subexpressions
  (APP    (subexprs list?))
  
  (BEGIN  (subexprs list?))

  (DEFINE (name ID?) (exp EXPR?))

  (SET!   (name ID?) (exp EXPR?))
  
  ; Args is a list of symbols that are the ids.
  (LAMBDA (args list?) (body EXPR?)))
  
 
; The value domain
(define-type VAL
  (NUM    (value number?))
  (BOOL   (value boolean?))
  (SYM    (value symbol?))
  (VOID)
  (PAIR   (first VAL?) (rest VAL?))
  (NULL)
  (PROC   (formals list?) (evt evt?) (body EXPR?))
  
  ; BUILTINS contain a Scheme procedure that takes a list of
  ; VALs and returns a VAL
  (BUILTIN-PROC (scheme-proc procedure?)))

; Simple value -> VAL
(define (->VAL v)
  (cond [(void? v)    (VOID)]
        [(symbol? v)  (SYM v)]
        [(boolean? v) (BOOL v)]
        [(number? v)  (NUM v)]
        [(null? v)    (NULL)]
        [else         (error '->VAL "Illegal Scheme type")]))

; Converts any value to a Scheme string
(define (->string v)
  (let ([o (open-output-string)])
        (print v o)
        (get-output-string o)))

(define (->Scheme e)
  (if (VAL? e)
      (cond [(BOOL? e) (BOOL-value e)]
            [(NUM? e)  (NUM-value e)]
            [(SYM? e)  (SYM-value e)]
            [(NULL? e) null]
            [(VOID? e) (void)]
            [else      (error 'Eph 
                              (string-append "Cannot convert Eph value \"" 
                                             (->string e) 
                                             "\" to a Scheme value"))])
      (error '->Scheme (string-append "\"" (->string e) "\" is not a VAL"))))

(define TRUE  (BOOL #t))
(define FALSE (BOOL #f))

; #t if v is TRUE
(define (TRUE? v) (and (BOOL? v) (BOOL-value v)))
