;; The first three lines of this file were inserted by DrScheme. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "reader.ss" "plai" "lang")
;; interpreter.ss
;;
;; Interpreter for the Eph language
;;
;; CS334: Principles of Programming Languages
;; http://cs.williams.edu/~morgan/www/cs334

(require "grammar.ss")
(require "parser.ss")
(require "evt.ss")

; eval: EXPR -> VAL
(define (eph-eval exp evt)
  (type-case EXPR exp
   [LIT (v)     
       ; Literal: convert directly to a VAL
        (->VAL v)]
   
   [IF (t c a)
       ; IF: process with a Scheme IF
       (eph-eval (if (TRUE? (eph-eval t evt)) c a) evt)]
   
   [VAR (name)
        ; VAR: look up variable in the environment
        (evt-get evt name)]
   
   [APP (subexprs)
        ; APP: first, eval subexpressions,
        (let ([values (map (λ (sub) (eph-eval sub evt)) subexprs)])
          ; ...and then apply the first to the rest
          (eph-apply (first values) (rest values)))]

   [BEGIN (subexprs)
        ; BEGIN: Eval subexpressions in order, then return the last
        (foldl (λ (sub ignore) (eph-eval sub evt)) (VOID) subexprs)]
   
   [DEFINE (name exp2)
        (if (evt-this-contains? evt name)
            ; Error check
            (variable-error "Multiple definition of " name)
            
            ; Normal case
            (begin
              ; First, bind the name to uninitialized in order
              ; to detect uses of it
              (evt-define! evt name UNINITIALIZED-VAR)
  
              ; Second, evaluate the expression and then bind the name to it
              (evt-set! evt name (eph-eval exp2 evt))
              
              ; Return the void value
              (VOID)))]
          
   [SET! (name exp2)
         (if (evt-initialized? evt name)
             ; Normal case: mutate [note that we are allowed to SET!
             ; in the middle of a DEFINE]
             (begin
               (evt-set! evt name (eph-eval exp2 evt))
               (VOID))
            
             ; Error case: in the middle of a define
             (variable-error 
              "Cannot set identifier before its definition" name))]
   
   [LAMBDA (args body)
           ; LAMBDA: create a new procedure. Do not evaluate the
           ; body.
           
           (begin
             ; Check the args
             (for-each 
              (λ (a) 
                (when (not (ID? a))
                    (variable-error "Not a legal identifier: " a))) args)
             
             (PROC args evt body))]))
 

; eph-apply : (PROC | BUILTIN-PROC) x list(VAL) -> VAL
(define (eph-apply proc actuals)
  (cond 
    [(BUILTIN-PROC? proc)
     ; Run the built in proc
     ((BUILTIN-PROC-scheme-proc proc) actuals)]
        
     
    [(PROC? proc)
     (eph-apply-PROC proc actuals)]
    
    [else  (error 'Eph (string-append "First subexpression in an APP"
                                      " must evaluate to a procedure"))]))

; eph-apply-proc : PROC x list(VAL) -> VAL
(define (eph-apply-PROC proc actuals)
  (let ([body    (PROC-body proc)]
        [formals (PROC-formals proc)]

        ; Create a new environment in which the procedure will execute.
        ; The parent of this environment is the one in which the proc
        ; was defined (lexical scope).
        [evt     (make-child-evt (PROC-evt proc))])

    (when (neq? (length formals) (length actuals))
      (error 'Eph "Wrong number of arguments in an APP"))
    
    ; Bind the formals to the actuals in the new environment
    (for-each (λ (id val) (evt-define! evt id val)) formals actuals)

    ; Execute the body in the new environment
    (eph-eval body evt)))

  