;; The first three lines of this file were inserted by DrScheme. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "reader.ss" "plai" "lang")
;; library.ss
;;
;; Built-in procedures for the Eph language.  This is the "library" 
;; environment
;;
;; CS334: Principles of Programming Languages
;; http://cs.williams.edu/~morgan/www/cs334

(require "grammar.ss")
(require "parser.ss")
(require "evt.ss")
(require "interpreter.ss")

; Wraps a Scheme procedure with an Eph procedure, by
; converting inputs and outputs between the languages.
(define (wrap scheme-proc)
  (BUILTIN-PROC
   (λ (eph-values)
     ; Convert result back to Eph value
     (->VAL
      (apply scheme-proc
             ; Convert eph values to Scheme values
             (map ->Scheme eph-values)))))) 

(define (make-library-evt)
  (let ([evt (make-root-evt)])
  
    ; Export Scheme procedures to Eph procedures using the
    ; same variable names
    
    (evt-define! evt '+ (wrap +))
    (evt-define! evt '- (wrap -))
    (evt-define! evt '* (wrap *))
    (evt-define! evt '/ (wrap /))
    (evt-define! evt 'remainder (wrap remainder))

    (evt-define! evt '> (wrap >))
    (evt-define! evt '< (wrap <))
    (evt-define! evt '>= (wrap >=))
    (evt-define! evt '<= (wrap <=))
    (evt-define! evt 'not (wrap not))

    ; Direct data accessors
    (evt-define! evt 'cons  (BUILTIN-PROC (λ (args) (PAIR (first args) (second args)))))
    (evt-define! evt 'first (BUILTIN-PROC (λ (args) (PAIR-first (first args)))))
    (evt-define! evt 'rest  (BUILTIN-PROC (λ (args) (PAIR-rest  (first args)))))
    (evt-define! evt 'list  (BUILTIN-PROC (λ (args) (foldl PAIR (NULL) args))))

    (evt-define! evt 'eq?   (BUILTIN-PROC eph-eq?))
    
    (evt-define! evt 'apply (BUILTIN-PROC (λ (args) (eph-apply (first args) 
                                                               (PAIR->list (second args))))))

    ; Define others as Eph procedures using the Eph 
    ; language itself (bootstrap)
    
    (eph-eval 
     (parse 
      '(begin
        
         (define second (λ (L) (first (rest L))))
         (define third  (λ (L) (first (first (rest L)))))
         (define zero?  (λ (n) (equals? 0 n)))
         (define null?  (λ (x) (eq? null x)))
         
         (define factorial 
           (λ (n) (if (eq? n 0) 1 (* (factorial (- n 1))))))         
        
        ) ; end begin
      ) ; end parse
     evt)
    
    evt))

; list(VAL) -> VAL
(define (eph-eq? args)
  (BOOL
   ; Recall that a and b have type VAL, not Scheme value
   (let ([a (first args)]
         [b (second args)])
     (cond
       ; Two procedures must be the same
       ; exact object to be eq?
       [(PROC? a) (eq? a b)]
       
       ; Void is only eq? to another void
       [(VOID? a) (VOID? b)]
       
       ; Null is only eq? to another null
       [(NULL? a) (NULL? b)]
       
       ; Everything else compares under Scheme's eq?
       [else (eq? (->Scheme a) (->Scheme b))]))))

; PAIR -> list(VAL)
(define (PAIR->list p)
  (if (NULL? p)
      '()
      (cons (PAIR-first p) (PAIR->list (PAIR-rest p)))))

