;; The first three lines of this file were inserted by DrScheme. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "reader.ss" "plai" "lang")
;; parser.ss
;;
;; Parser for the Eph language
;;
;; CS334: Principles of Programming Languages
;; http://cs.williams.edu/~morgan/www/cs334

(require "grammar.ss")

; Scheme literal | list -> EXPR
(define (parse e)
  (cond
    ; Note the number and boolean literals do *not* become NUM and BOOL,
    ; because those are in the value domain.  Instead, they become literal
    ; expressions.
    [(number? e)  (LIT e)]
    [(boolean? e) (LIT e)]
    
    ; For convenience, we allow these special symbols
    [(eq? 'true e)  TRUE]
    [(eq? 'false e) FALSE]
     
    ; Note: in PLT Scheme, null = empty = '()
    [(eq? 'null e)  (LIT null)]

    ; Note: void is a literal in Eph; in Scheme it is a procedure
    [(eq? 'void e)  (LIT (void))]
    
    ; Raw symbols are Eph variables (*quoted* symbols 
    ; are Eph symbols; see below in the list case)
    [(ID? e)      (VAR e)]
        
    ; If the expression looks like ( ... ), choose the type
    ; based on what is first.
        
    [(list? e)
          
     (case (first e)
       ; This is a symbol.  As with numbers and booleans, this becomes
       ; a LIT, not a SYM
       [(quote)    (LIT (second e))]   
       
       ; (lambda (args) body)
       [(lambda λ) (LAMBDA (second e) (parse (third e)))]
        
       ; (if t c a)
       [(if) (IF (parse (second e)) (parse (third e)) (parse (fourth e)))]

       ; (and a b)
       ; Rewrite as an IF
       [(and) (IF (parse (second e)) (parse (third e)) (LIT #f))]

       ; (or a b)
       ; Rewrite as an IF
       [(or) (IF (parse (second e)) (LIT #t) (parse (third e)))]
       
       ; (begin expr1 expr2 ... )
       [(begin) (BEGIN (map parse (rest e)))]
       
       ; (define id expr)
       [(define) (DEFINE (second e) (parse (third e)))]
       
       ; (set! id expr)
       [(set!) (SET! (second e) (parse (third e)))]
       
       ; Anything else must be an APP
       ; (expr1 expr2 ...)
       (else     (APP (map parse e))))]
    
    [else  (error 'parse (string-append "Illegal Eph expression: " (->string e)))]))
