;; The first three lines of this file were inserted by DrScheme. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "reader.ss" "plai" "lang")
;; test.ss
;;
;; Tests of the Eph interpreter
;;
;; CS334: Principles of Programming Languages
;; http://cs.williams.edu/~morgan/www/cs334

(require "interpreter.ss")
(require "parser.ss")
(require "grammar.ss")
(require "library.ss")

; (assert bool)
; (assert bool message-1 ... message-n)
; Prints the messages if b is false
(define (assert b . m)
  (if (not b)
      (error 'assert 
             (if (empty? m)
                 "Assertion Failed"
                 (map string-append (cons "Assertion Failed: " m))))
      (void)))

; Tests the source [which is unparsed] and returns the result as a VAL
(define (test source) 
  (display "_________________________________________________")
  (newline)
  (newline)
  (display "Source:     ")
  (display source)
  (newline)
  
  (let ([expression (parse source)])
    
    (display "Parsed:     ")
    (print expression)
    (newline)
    
    (let ([result (eph-eval expression (make-library-evt))])
    
      (display "Result:     ")
      (print result)
      (newline)
  
      result)))

; LIT
(assert (NUM?  (test 1)))
(assert (BOOL? (test #f)))
(assert (SYM?  (test ''x)))
(assert (NULL? (test 'null)))
(assert (VOID? (test 'void)))

; VAR
(assert (BUILTIN-PROC? (test '+)))

; APP
(assert (eq? 3 (->Scheme (test '(+ 1 2)))))

; IF
(assert (eq? 'x (->Scheme (test '(if #t 'x 'y)))))
(assert (eq? 'y (->Scheme (test '(if #f 'x 'y)))))

; AND
(assert (eq? #f (->Scheme (test '(and #t #f)))))
(assert (eq? #t (->Scheme (test '(and #t #t)))))

; OR
(assert (eq? #f (->Scheme (test '(or #f #f)))))
(assert (eq? #t (->Scheme (test '(or #f #t)))))
(assert (eq? #t (->Scheme (test '(or #t #t)))))

; DEFINE + BEGIN
(assert (eq? 3 (->Scheme (test '(begin (define x 3) x)))))

; LAMBDA
(assert (eq? 0 (->Scheme (test '((λ (x) x) 0)))))

; LAMBDA with recursive environment lookup
(assert (BUILTIN-PROC? (test '((λ () +)))))

; LAMBDA with APP
(assert (eq? 0 (->Scheme (test '((λ (x) (- x 1)) 1)))))

; SET!
(assert (eq? 0 (->Scheme (test '(begin
                                  (define x 4)
                                  (set! x 1)
                                  (set! x (- x 1))
                                  x)))))
                                      

; Sample recursive procedure
(assert (eq? 6 (->Scheme (test 
                          '(begin
                             (define sum (λ (L)
                                           (if (null? L)
                                               0
                                               (+ (first L) (sum (rest L))))))
                             
                             (define my-list (list 1 2 3))
                             
                             (sum my-list))))))

; Shadowing 
(assert (eq? 4 (->Scheme (test
                          '(begin
                             (define x 4)
                             ((λ (x)
                                (set! x 3)
                                x) 2)
                             x)))))

; Lexical scope
(assert (eq? 4 (->Scheme (test
                          '(begin
                             (define x 4)
                             (define f (λ () x))
                             ((λ (x) (f)) 10))))))

; Apply
(assert (eq? 3 (->Scheme (test '(apply + (list 1 2))))))
